
#include "guis/gnome/decoder_gnome.hh"
#include "input/streamsrc_istr.hh"
#include "output/out_x11.hh"

#include "system/sysdec1.hh"
#include "system/resync.hh"
#include "system/sysdemux.hh"
#include "system/videoauconv.hh"
#include "system/syncer.hh"
#include "system/ptsvideo.hh"
#include "video12/vdecoder.hh"

#include <libvideogfx/libvideogfx.hh>

#include <iostream>
#include <fstream>
using namespace std;

extern GtkWidget* drawingarea;

// stream sources

//static StreamSource* streamsrc;
static StreamSource_IStream strsrc_istr;

// the system decoder

static SystemDecoder_MPEG1 system_dec;
static SystemResyncer      system_resync;
static SystemDemux         system_demux(system_resync);

// video decoding pipeline

static SystemDemuxExtractor video_extractor;
static VideoAUConverter     vidcon(video_extractor);
//static PTSVideoCompleter    video_ptsvideo(vidcon);
static VideoDecoder         video_dec(vidcon); //video_ptsvideo);

// synchronization

static Syncer_Realtime syncer_realtime;
static Syncer* syncer = &syncer_realtime;

// input configuration

static ifstream istr;

static void ViewMPEG();


// ----------- Gtk output class ---------------

static bool Image2GdkRGBBuf(const Image_YUV<Pixel>& img,guchar** buf,int* width,int* height)
{
  bool alloc_buf = false;

  if (img.AskWidth() != *width ||
      img.AskHeight() != *height ||
      *buf == NULL)
    alloc_buf=true;

  if (alloc_buf)
    {
      *width  = img.AskWidth();
      *height = img.AskHeight();

      if (*buf) delete[] *buf;

      *buf = new guchar[*width * *height * 3];
    }

  RawImageSpec_RGB spec;
  spec.bytes_per_line = (*width)*3;
  spec.bits_per_pixel = 24;
  spec.little_endian  = false;
  spec.SetRGBMasks(0xFF0000,0x00FF00,0x0000FF);

  Image2Raw transformation;
  transformation.SetOutputSpec(spec);
  transformation.TransformYUV(img , (uint8*)(*buf));

  return alloc_buf;
}




class VideoSink_Gtk : public VideoOutput,
		      public ImageDataAccumulator
{
public:
   VideoSink_Gtk();
  ~VideoSink_Gtk();

  void SetWidget(GtkWidget* drawingarea);

  void ShowMBRows(DecodedImageData*);
  void FinishedPicture();

  bool PictureAvailable() { return pic_available; }
  PTS  AskPTSOfNextToBeDisplayed() const { return nextpts; }
  void ShowPicture() { pic_available=false; }
  void Reset();

  void Redraw();

private:
  //GtkWidget* d_drawingarea;
  guchar*    d_buf;
  int d_buf_width,d_buf_height;

  bool first;
  int  d_height;

  bool pic_available;
  PTS  nextpts;

  GdkGC* gc;
};


VideoSink_Gtk::VideoSink_Gtk()
{
  Reset();
}

void VideoSink_Gtk::Reset()
{
  d_buf=NULL;
  d_buf_width=0;
  d_buf_height=0;
  first=true;
  pic_available=false;
  gc=NULL;
}

void VideoSink_Gtk::SetWidget(GtkWidget* drawingarea)
{
  //d_drawingarea=drawingarea;
  gc = gdk_gc_new(drawingarea->window);
}

VideoSink_Gtk::~VideoSink_Gtk()
{
}

void VideoSink_Gtk::ShowMBRows(DecodedImageData* dimg)
{
  const Image_YUV<Pixel>& img = dimg->m_image;

  nextpts = dimg->m_timing.pts;

  if (first)
    {
      ImageParam_YUV param;
      img.GetParam(param);
      d_height = param.height;

      gtk_drawing_area_size(GTK_DRAWING_AREA(drawingarea),param.width,param.height);

      StartAccumulation(0,d_height-1,false);
  
      first=false;
    }

  DecodedImageData* decimg = Accumulate(dimg);
  if (!decimg)
    return;

  Image2GdkRGBBuf(decimg->m_image,&d_buf,&d_buf_width,&d_buf_height);

#if 1
  gdk_draw_rgb_image(drawingarea->window,gc,
		     0,0,d_buf_width,d_buf_height,
		     GDK_RGB_DITHER_NORMAL,
		     d_buf,d_buf_width*3);
#endif
}


void VideoSink_Gtk::Redraw()
{
  gdk_draw_rgb_image(drawingarea->window,gc,
		     0,0,d_buf_width,d_buf_height,
		     GDK_RGB_DITHER_NORMAL,
		     d_buf,d_buf_width*3);
}


void VideoSink_Gtk::FinishedPicture()
{
  pic_available=true;
  StartAccumulation(0,d_height-1,false);
}



// video output

static VideoSink_X11 videosink_x11;
static VideoSink_Gtk videosink_gtk;
static VideoOutput* outputobject = &videosink_gtk;

void Gnome_Redraw()
{
  videosink_gtk.Redraw();
}

// --------------------------------------------------------------


void Gnome_DecoderInit()
{
  system_resync.SetSource(&system_dec);

  system_demux.EnableChannel(StreamIDBase_Video|0);

  video_extractor.SetSource(&system_demux);
  video_extractor.SetChannel(StreamIDBase_Video|0);

  video_dec.SetSink(*outputobject);
}

static void DecoderReset()
{
  system_dec.NewFileReset();
  system_demux.Reset();
  vidcon.Reset();
  video_dec.Reset();

  VideoDecoder_Options opts;

  opts.Trace_SeqH=
    opts.Trace_GOPH=
    opts.Trace_PicH=
    opts.Trace_SlcH=false;
  opts.Trace_MB=
    opts.Trace_SliceData=
    opts.Trace_DCTCoeff=false;

    video_dec.SetOptions(opts);
}

bool Gnome_LoadMPEG(gchar* selected_filename)
{
  videosink_gtk.Reset();
  videosink_x11.Reset();

  istr.close();
  istr.open(selected_filename, std::ios::in | std::ios::binary);
  strsrc_istr.SetIStream(istr);

  system_dec.SetSource(strsrc_istr);

  videosink_gtk.SetWidget(drawingarea);
  DecoderReset();

  return true;
}

bool Gnome_ViewMPEG()
{
  try
    {
#if 0
      SysPacket* spck = system_dec.GetNextPacket();
      if (spck)
	{
	  cout << "pos: " << spck->absolute_filepos
	       << " pcknr: " << spck->pck_nr
	       << endl;
	  return true;
	}
      else
	{
	  cout << "EOF\n";
	  return false;
	}
#endif

#if 0
      SysPacket_Packet* pck = vidcon.GetNextPacket();

      if (pck)
	cout << "pos: " << pck->absolute_filepos
	     << " off: " << pck->data_start_offset
	     << " pcknr: " << pck->pck_nr
	     << " ch: " << ((int)pck->Channel)
	     << " len: " << pck->data.AskLength()
	     << endl;
      else
	{
	  cout << "EOF\n";
	  return false;
	}
#endif

#if 1
      if (outputobject->PictureAvailable())
	{
#if 0
	  if (outputobject->AskPTSOfNextToBeDisplayed())
	    syncer->WaitUntil(outputobject->AskPTSOfNextToBeDisplayed());
#endif
	  outputobject->ShowPicture();
	}
      else
	{
	  bool more = video_dec.DecodeAFrame();
	  if (!more)
	    return false;
	}
#endif

      return true;
    }
  catch(...)
    {
      cout << "Unexpected end.\n";
      return false;
    }
}




#if 0

#include "error.hh"
#include "libvideogfx/init.hh"
#include "config.h"

#include "input/streamsrc_cin.hh"
#include "input/streamsrc_udp.hh"
#include "input/errors.hh"
#include "vpostproc/buffer.hh"

#include "vpostproc/pp_resize.hh"
#include "vpostproc/pp_filteredresize.hh"
#include "vpostproc/pp_mblks.hh"
#include "vpostproc/pp_mv.hh"
#include "vpostproc/pp_mvcol.hh"
#include "vpostproc/pp_qscale.hh"
#include "vpostproc/pp_fps.hh"
#include "output/out_ppm.hh"
#include "output/out_yuv.hh"
#include "version.hh"

#if LINUX
#include "output/out_mgavid.hh"
#include "input/streamsrc_linux_vcd.hh"
#endif

#include "libvideogfx/x11/imgwin.hh"
#include "libvideogfx/x11/server.hh"


#include "options.hh"


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#ifdef _WIN32
#include <fcntl.h>
#include <io.h>
#endif

#if 0
#include "libvideogfx/arch/x86/CPUcapabilities.hh"
#endif


static StreamSource_cin     strsrc_cin;
static StreamSource_udp     strsrc_udp;
#if LINUX
static StreamSource_VCD     strsrc_vcd;
#endif


// audio decoding pipeline

static SystemDemuxExtractor audio_extractor;
static SystemDemuxExtractor ac3_extractor;
static int ac3stream = 0x80;

// video postprocessors

static DecodedPictureSink* lastoutput;
static Resize_Interface*   resizer;
static VideoPostprocessor_MBBoundaries   pp_mblks;
static VideoPostprocessor_MotionVector   pp_mv;
static VideoPostprocessor_MVCol          pp_mvcol;
static VideoPostprocessor_QScale         pp_qscale;
static VideoPostprocessor_FramesPerSec   pp_fps;
static VideoPostprocessor_Resize         pp_resize;
static VideoPostprocessor_FilteredResize pp_filteredresize;
#if 0
static VideoPostprocessor_ImageType    pp_imgtype;
#endif


// video output

static VideoSink_YUVFile videosink_yuv;
static VideoSink_PPMFile videosink_ppm;
#if LINUX
static VideoSink_MGA videosink_mga;
#endif
#if DIRECTFB_SUPPORT
static VideoSink_DirectFB videosink_directfb;
#endif


void InitStreamSource(ifstream& istr)
{
  strsrc_istr.SetIStream(istr);
  streamsrc=&strsrc_istr;
}

void InitStreamSource_cin()
{
  streamsrc=&strsrc_cin;
}


void InitStreamSourceVCD(int track)
{
#if LINUX
  strsrc_vcd.Init();
  if (track<0 || track>strsrc_vcd.AskNTracks())
    throw "VCD track out of range";

  strsrc_vcd.SkipToTrack(track);
  streamsrc=&strsrc_vcd;
#endif
}


void InitAudioDecoder()
{
  system_demux.EnableChannel(StreamIDBase_Audio|0);
  audio_extractor.SetSource(&system_demux);
  audio_extractor.SetChannel(StreamIDBase_Audio|0);
}


void InitAC3Decoder()
{
  system_demux.EnableChannel(0xBD);
  ac3_extractor.SetSource(&system_demux);
  ac3_extractor.SetChannel(0xBD);
}


void InsertPostprocessor(VideoPostprocessor& pp)
{
  pp.SetNext(lastoutput);
  lastoutput = &pp;
}




int maxframes;


static void usage()
{
  cerr <<
    " DVDview " DVDVIEW_VERSION "    " DVDVIEW_DATE " (c) Dirk Farin\n"
    "------------------------------------------------\n"
    "usage: dvdview <options> filename.mpg\n"
    "\n"
    "Options:\n"
    "  -B    Skip B-frames\n"
    "  -P    Skip P-frames (and B-frames)\n"
    "  -T #  Set speed (100 for realtime, >100 for faster)\n"
    "  -S #  Set output size.\n"
    "  -N #  Number of frames to decode\n"
    "  -v #  Enable logging level (0 <= # <= 7)\n"
    "         0 - off\n"
    "         1 - sequence headers\n"
    "         2 - gop headers\n"
    "         3 - picture headers\n"
    "         4 - slice headers\n"
    "         5 - macroblocks\n"
    "         6 - DCT coefficients\n"
    "         7 - binary slice data\n"
    "  -F    Write PPM sequence rather than display stream contents.\n"
    "  -Y    Write single YUV file rather than display stream contents.\n"
    "  -W s  Write YUV with additional header information to file 's'.\n"
    "  -U s  Bind DVDview input to a UDP port.\n"
    "  -M    Show macroblock boundaries (twice for bigger marks).\n"
    "  -d      Show boundary as dark pixels.\n"
    "  -l      Show boundary as bright pixels.\n"    
    "  -Q    Show QScale of macroblocks.\n"
    "  -V    Show motionvectors of macroblocks (vectors).\n"
    "  -C    Show motionvectors of macroblocks (colors).\n"
    "  -p      Enable vectors in P-frames.\n"
    "  -b      Enable vectors in B-frames.\n"
    "  -c      Show colored motionvectors.\n"
    "  -f      Enable forward vectors.\n"
    "  -r      Enable backward vectors.\n"
    "  -O      hold mode\n"
    "  -s    Show decoder speed in fps.\n"
    "  -A    Extract audio-stream to stdout. Pipe this into mpg123.\n"
    "  -3    Extract AC3-stream to stdout. Pipe this into ac3dec.\n"
    "  -a #    select AC3-stream (0-7) default is 0.\n"
#if 0
    "  -W    Set time to wait after each frame (in 1/100s).\n"
    "  -S    Disable X11 shared memory extension.\n"
    "  -F    DisableSomething=true (development only).\n"
    "  -m    Mark special MBs (development only).\n"
#endif
    "  -h    Show this usage information\n"

#if LINUX
    " - - - - - -  Linux only features  - - - - - -\n"
    "  -D #  Play VideoCD track rather that MPEG file (omit filename)\n"
    "  -X    Use X11 output even when MGA_VID is available.\n"
#endif
    ;
}


extern void GetLinesForDrawing(DecodedImageData* dimg,int first,int last);

#if 0
int main(int argc,char** argv)
{
  InitULib(argc,argv);

  DecodedImageData dimg;
  dimg.m_may_modify=false;
  dimg.m_field_lines=true;
  dimg.m_src_y_start=10;
  dimg.m_src_y_end  =40;
  dimg.m_dst_y_start=1;
  GetLinesForDrawing(&dimg,10,20);
}
#endif

void start_GUI_shell(int argc,char** argv)
{
#ifdef _WIN32
  /* Switch Win32 standard output to binary mode, so that piping the
     raw image data works. */
  {
    int result = _setmode( _fileno(stdout), _O_BINARY );
    if (result == -1)
      {
	cerr << "Cannot set standard output to binary mode.\n";
	exit(10);
      }
  }
#endif

  VideoDecoder_Options vdec_options;

  bool imgtypes=false;
  bool qscales=false;
  bool mvs=false,mvcol=false;
  bool audio=false,ac3=false;
  bool mvf=false,mvb=false,mvcolor=false;
  bool mvpf=false,mvbf=false;
  bool mv_holdmode=false;
  bool fpsdisplay=false;
  bool x11_output=false;   // try MGA first
  int vcd=0;
  bool resize=false;
  int scaled_width,scaled_height;
  bool rescale=false;
  bool filtered_resize=false;
  bool udp=false;

#if 0 && ENABLE_MMX
  cout << "CPU capabilities detected: ";
  if (cpucapabilities.HasMMX()) cout << "MMX ";
  if (cpucapabilities.HasKNI()) cout << "KNI ";
  if (cpucapabilities.Has3dNow()) cout << "3dNow! ";
  if (cpucapabilities.HasMTRR()) cout << "MTRR ";
  if (cpucapabilities.HasCMOV()) cout << "CMOV ";
  if (cpucapabilities.HasFPU()) cout << "FPU ";
  cout << endl;
#endif

#if LINUX
  // Try to use MGA if available.

  bool mga_available = videosink_mga.MGA_Available();
  if (mga_available)
    {
      outputobject = &videosink_mga;
      resizer = &videosink_mga;
    }
  else
    {
      outputobject = &videosink_x11;
      resizer = &pp_resize;
    }
#else
  // Use X11 output by default

  outputobject = &videosink_x11;
  resizer = &pp_resize;

  if (!default_x11server.AskDisplay())
    { cerr << "Cannot open display.\n"; throw "no display"; }
#endif

  try
    {
#if 1
      int c;

      while ((c=getopt(argc,argv,"qf3rlcdv:wpuT:S:BbPHMQVW:FYOCdmAa:hXxD:sN:U:")) != EOF)
	{
	  switch (c)
	    {
	    case 'v':
	      {
#ifndef NDEBUG
		int level=atoi(optarg);
		if (level>7) level=7;
		if (level<0) level=0;
		switch(level)
		  {
		  case 7: vdec_options.Trace_SliceData = true;
		  case 6: vdec_options.Trace_DCTCoeff  = true;
		  case 5: vdec_options.Trace_MB   = true;
		  case 4: vdec_options.Trace_SlcH = true;
		  case 3: vdec_options.Trace_PicH = true;
		  case 2: vdec_options.Trace_GOPH = true;
		  case 1: vdec_options.Trace_SeqH = true;
		  case 0: break;
		  }
#else
		cerr << "Ignoring option -v#: Verbose output was not compiled in.\n";
#endif
	      }
	      break;
	      
	    case 'M':
	      if (!options.Postproc_ShowMBs)
		options.Postproc_ShowMBs = true;
	      else
		{ pp_mblks.SetBigMarks(); }
	      break;

	    case 'd':
	      pp_mblks.SetStaticMode(0);
	      break;

	    case 'l':
	      pp_mblks.SetStaticMode(255);
	      break;

	    case 'P':
	      vdec_options.DecodeP = vdec_options.DecodeB = false;
	      break;
	      
	    case 'B':
	      vdec_options.DecodeB = false;
	      break;

	    case 'x':
	      imgtypes=true;
	      break;
	      
	    case 'Q':
	      qscales=true;
	      break;
	      
	    case 'V':
	      mvs=true;
	      break;

	    case 'C':
	      mvcol=true;
	      break;

	    case 'O':
	      mv_holdmode=true;
	      break;

	    case 'c': mvcolor=true; break;
	    case 'f': mvf=true; break;
	    case 'r': mvb=true; break;
	    case 'p': mvpf=true; break;
	    case 'b': mvbf=true; break;
	      
	    case 'A':
	      audio=true;
	      break;
	      
	    case '3':
	      ac3=true;
	      break;

	    case 'a':
	      ac3stream = atoi(optarg) + 0x80;
	      break;

	    case 'N':
	      maxframes = atoi(optarg);
	      break;

#if DIRECTFB_SUPPORT
	    case 'q':
	      outputobject = &videosink_directfb;
	      resizer = &pp_resize;
	      break;
#endif

	    case 'F':
	      outputobject = &videosink_ppm;
	      resizer = &pp_resize;
	      break;

	    case 'Y':
	      outputobject = &videosink_yuv;
	      resizer = &pp_resize;
	      break;

	    case 'W':
	      outputobject = &videosink_yuv;
	      resizer = &pp_resize;
	      videosink_yuv.SetOutputFile(optarg);
	      videosink_yuv.WriteHeaders(true);
	      break;

	    case 's':
	      fpsdisplay=true;
	      break;

	    case 'h':
	      usage();
              exit(10);
	      break;

	    case 'T':
	      {
		int speed=atoi(optarg);
		if (speed<=0)
		  {
		    cerr << "Speed must be greater than 0%, using 1%\n";
		    speed=1;
		  }
		syncer_realtime.SetSpeed(speed);
	      }
	      break;

	    case 'S':
	      {
		char* w = strtok(optarg,"xX");
		char* h = strtok(NULL,"xX");
		if (!w || !h)
		  { cerr << "invalid size argument (1)\n"; exit(5); }
		
		scaled_width  = atoi(w);
		scaled_height = atoi(h);

		rescale=true;
	      }
	      break;

	    case 'H':
	      filtered_resize = true;
	      break;

	    case 'X':
	      x11_output=true;
	      break;

	    case 'U':
	      strsrc_udp.SetUDPEndpoint("rachmaninoff.informatik.uni-mannheim.de",1234);
	      streamsrc=&strsrc_udp;
	      udp=true;
	      break;

#if LINUX
	    case 'D':
	      vcd=atoi(optarg);
	      break;
#endif
	    }
	}

      if (x11_output)
	{
	  outputobject = &videosink_x11;
	  resizer = &pp_resize;

	  if (!default_x11server.AskDisplay())
	    { cerr << "Cannot open display.\n"; throw "no display"; }
	}

      if (filtered_resize && resizer==&pp_resize) { cout << "bla\n"; resizer=&pp_filteredresize; }

      lastoutput = outputobject;


      if (optind==argc && !vcd && !udp)
	{ usage(); exit(5); }

      video_dec.SetOptions(vdec_options);

      if (rescale) resizer->SetScaledSize(scaled_width,scaled_height);

      ifstream str;

      if (vcd)
	{
	  InitStreamSourceVCD(vcd);
	}
      else if (!udp)
	{
	  if (strcmp(argv[optind],"-")==0)
	    InitStreamSource_cin();
	  else
	    {
	      str.open(argv[optind], std::ios::in | std::ios::binary);
	      InitStreamSource(str);
	    }
	}

      InitVideoDecoder1();
      if (audio) InitAudioDecoder();
      if (ac3)   InitAC3Decoder();

      if (filtered_resize) InsertPostprocessor(pp_filteredresize); else InsertPostprocessor(pp_resize);
      if (qscales)  InsertPostprocessor(pp_qscale);
      if (options.Postproc_ShowMBs) InsertPostprocessor(pp_mblks);
      if (fpsdisplay) InsertPostprocessor(pp_fps);
      if (mvs)
	{
	  if (!mvf && !mvb) { mvf=mvb=true; }
	  pp_mv.SelectMVs(mvf,mvb);
	  pp_mv.ColoredVectors(mvcolor);
	  pp_mv.SelectFrametypes(mvpf,mvbf);
	  InsertPostprocessor(pp_mv);
	}
#if 0
      if (imgtypes) InsertPostprocessor(pp_imgtype);
#endif
      if (mvcol)
	{
	  pp_mvcol.SetHoldMode(mv_holdmode);
	  pp_mvcol.SelectFrametypes(mvpf,mvbf);
	  InsertPostprocessor(pp_mvcol);
	}
      InitVideoDecoder2();

      //str.seekg(200000,ios::beg);

      ViewMPEG(audio,ac3);
#endif

#if 0
      ifstream str(argv[1]);
      ShowSystemPackets(str);
#endif

#if 0
      ifstream str(argv[1]);
      //ShowVideoAccessUnitPackets(str);
      //ShowVideoSyntax(str);
      DecodeVideo(str);
#endif

#if 0
      CheckMemAlloc();
#endif

#if 0
      if (!ProcessOptions(argc,argv, options))
	return 0;
#endif
    }
  catch (Excpt_Base& e)
    {
      MessageDisplay::Show(e.m_severity , e.m_text);
    }
  catch (const char* txt)
    {
      cerr << "error: " << txt << endl;
    }
  catch (...)
    {
      cerr << "error occurred... quitting...\n";
    }
}


#endif
